/*
 * i2c.c - I2C API for C64
 *
 * Copyright (c) 2013 Timo Voutilainen
 *
 * This file is part of FM Radio for Commodore 64.
 *
 * FM Radio for Commodore 64 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FM Radio for Commodore 64 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FM Radio for Commodore 64.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "i2c.h"

//#define DEBUG	1

#define SDA_MASK 0x01 /* USR0 is SDA */
#define SCL_MASK 0x02
#define REST_OF_USRPORT_PINS_MASK 0xFC /* 11111100 b */
#define SLAVE_ACK_WAIT_TIME 50 /* How long we wait if the slave acks the address*/
#define USRPORT 0xDD01 /* 56577 */
#define USRPDIR 0xDD03 /* 56579, bit=1=Output */
#define OUTPUT 1 /* Userport direction */
#define INPUT 0 /* Userport direction */

/* TODO: Get rid of these globals - not needed necessarily */
static u8 SDA = HIGH;
static u8 SCL = HIGH;
static u8 sda_dir = OUTPUT;
static u8 scl_dir = OUTPUT;

static void set_sda(u8 state);
static void set_scl(u8 state);
static u8 read_sda();
static void write_start_bit();
static u8 i2c_write_a_byte(u8 bytetowrite, u8 rw_bit);
static u8 i2c_read_byte(void);

/* --------------------------------------------------------------- */
/* This one checks if all is ok to start */
u8 init_i2c_bus()
    {
    u8 status = OK;
#ifdef VICE
    return OK;
#endif

    POKE(USRPDIR,((sda_dir*SDA_MASK)|(scl_dir*SCL_MASK)));
    set_sda(HIGH);
    set_scl(HIGH);
    time_delay();
    sda_dir = INPUT;
    scl_dir = INPUT;
    POKE(USRPDIR,((sda_dir*SDA_MASK)|(scl_dir*SCL_MASK)));
    if (PEEK(USRPORT) != 255)
        {
        printf("Warning - line is being pulled down!\n");
        status = NOT_OK;
        }
    sda_dir = OUTPUT;
    scl_dir = OUTPUT;
    POKE(USRPDIR,((sda_dir*SDA_MASK)|(scl_dir*SCL_MASK)));
    set_sda(HIGH);
    set_scl(HIGH);

    //Send stop just in case...
    write_stop_bit();
    return (status);
    }

/* --------------------------------------------------------------- */
void time_delay()
    {
    /* You may implement your own timer based delays etc. here */
    /*int i=0;

    while (i<1000)
       {
       i++;
       }*/
    }
/* --------------------------------------------------------------- */

static void set_sda(u8 state)
    {
    SDA = state;
    sda_dir = OUTPUT;
    POKE(USRPDIR,((sda_dir*SDA_MASK)|(scl_dir*SCL_MASK)));
    POKE(USRPORT,((REST_OF_USRPORT_PINS_MASK|(SDA*SDA_MASK)|(SCL*SCL_MASK))));
    }
/* --------------------------------------------------------------- */
static void set_scl(u8 state)
    {
    SCL = state;
    scl_dir = OUTPUT;
    POKE(USRPDIR,((sda_dir*SDA_MASK)|(scl_dir*SCL_MASK)));
    POKE(USRPORT,((REST_OF_USRPORT_PINS_MASK|(SDA*SDA_MASK)|(SCL*SCL_MASK))));
    }
/* --------------------------------------------------------------- */
static u8 read_sda()
    {
    u8 read_sda;
    sda_dir = INPUT;
    POKE(USRPDIR,((sda_dir*SDA_MASK)|(scl_dir*SCL_MASK)));
    read_sda = (PEEK(USRPORT)&SDA_MASK);
    return(read_sda);
    }
/* --------------------------------------------------------------- */
void write_start_bit()
    {
    set_sda(HIGH);
    set_scl(HIGH);    /* Make sure that SDA and SCL are High*/
    time_delay();
    set_sda(LOW);   /* First pull down SDA while SCL is still high */
    time_delay(); /* Then pull down SCL also  */
    set_scl(LOW);
    }
/* --------------------------------------------------------------- */
u8 write_stop_bit()
    {
    set_sda(LOW);
    time_delay();
    set_scl(HIGH);
    time_delay();
    set_sda(HIGH);
    time_delay();
    /* See if SDA really went up or if slave keeps SDA low ...=Error situation */
    if (read_sda() == 0) return (NOT_OK);
    else return (OK);
    }
/* --------------------------------------------------------------- */
void write_master_ack()
    {
    set_sda(LOW);
    set_scl(HIGH);
    time_delay();
    set_scl(LOW);
    time_delay();
    /* Notice that SDA and SCL is left low !!! */
    }
/* --------------------------------------------------------------- */
void write_master_nack()
    {
    set_sda(HIGH);
    set_scl(HIGH);
    time_delay();
    set_scl(LOW);
    time_delay();
    set_sda(LOW);
    /* Notice that SDA and SCL is left low !!! */
    }
/* --------------------------------------------------------------- */
u8 i2c_read_byte(void)
    {
    u8 read_byte=0;
    u8 bitcount = 8;

    while (bitcount > 0)
       {
       time_delay();
       set_scl(HIGH);
       time_delay();
       if(read_sda() == HIGH)
           read_byte = read_byte | 1;

       if (bitcount > 1 )
           read_byte = read_byte << 1;

       set_scl(LOW);
       time_delay();
       bitcount --;
       }
    return (read_byte);
    }
/* --------------------------------------------------------------- */
u8 i2c_write_a_byte(u8 bytetowrite, u8 rw_bit)
    {
    /*
    0000000 0 General Call
    0000000 1 Start Byte
    0000001 X CBUS Addresses
    0000010 X Reserved for Different Bus Formats
    0000011 X Reserved for future purposes
    00001XX X High-Speed Master Code
    11110XX X 10-bit Slave Addressing
    11111XX X Reserved for future purposes

    ==> Address is 7-bits long. List of reserved addresses:
    0,1,2,3,4,5,6,7,0x78,0x79,0x7A,0x7B,0x7C,0x7D,0x7E,0x7F
    */

    u8 wait_time = SLAVE_ACK_WAIT_TIME;
    u8 bytetemp;
    u8 bitcount = 8;
#ifdef DEBUG
	printf("i2c_wr_a_b:bytetow=0x%x, rw_bit=%d\n", bytetowrite, rw_bit);
#endif
    /* If this function should write an address then R/W bit needs to be added,
    otherwise do not shift and no r/w bit to be added */
    if (rw_bit <= READBIT) /* READBIT(1) or WRITEBIT(0) */
        {
        bytetowrite = bytetowrite << 1;
        bytetowrite = bytetowrite | rw_bit;
        }
    /* write_start_bit(); This is nolonger done here but needs to be done prior to this function */
    while (bitcount >0)          /* Until all 8 bits have been written */
        {
        bytetemp = bytetowrite;
        if (bytetemp & 128)   /* IF MSB is 1 then write 1 to bus ...*/
            {set_sda(HIGH);}
        else
            {set_sda(LOW);}  /* ... otherwise write 0     */

        time_delay();
        set_scl(HIGH);
        time_delay();
        set_scl(LOW);
        bytetowrite = bytetowrite << 1;
        bitcount -- ;
        }
    set_sda(HIGH);
    time_delay();
    set_scl(HIGH);

    while ( (read_sda() == HIGH) && (wait_time > 1) )
       {
        wait_time --;   /* Slave should acknowledge here = pull down the SDA */
        time_delay();
       }
    if ((wait_time <= 1) && (read_sda() == HIGH))
        {
        /* Error situation */
        return (NOT_OK);
        }
    else
        {
        set_scl(LOW);
        set_sda(HIGH);
        return (OK);
        }
    }
/* --------------------------------------------------------------- */
u8 i2c_read_register(u8 i2addr, u8 reg, u8 *hibyte, u8 *lobyte)
    {
	/* Reads a 16bit register  */
    u8 bitindex=8;

	write_start_bit();
	i2c_write_a_byte(i2addr, WRITEBIT); //Address and write -bit
	i2c_write_a_byte(reg, PURE_BYTE);
	write_start_bit();
	i2c_write_a_byte(i2addr, READBIT); //Address and read -bit
    //set_sda(1); Not needed ?
    *hibyte = i2c_read_byte();     /* Read the first byte */
    write_master_ack();
    time_delay();
    set_sda(HIGH);
    *lobyte = i2c_read_byte();      /* Read the second byte */
    write_master_nack();
#ifdef DEBUG
    printf("i2c_read_reg:reg=0x%x,hi=0x%x,lo=0x%x\n", reg, *hibyte, *lobyte);
#endif
    if (write_stop_bit() == NOT_OK)   /* And finally Stop */
        {
#ifdef DEBUG
	printf("i2c_read_register:write_stop failed!\n");
#endif
    	return NOT_OK;
        }
#ifdef DEBUG
    printf("highbyte : 0x%x, lobyte : 0x%x \n",hibyte, lobyte);
#endif
    return OK;
    }

/* --------------------------------------------------------------- */
u8 i2c_write_register(u8 i2caddr, u8 reg, u8 hibyte, u8 lobyte)
    {

  /* Writes a 16bit register  */
#ifdef DEBUG
	printf("i2c_write_reg:reg=0x%x,hi=0x%x,lo=0x%x\n", reg, hibyte, lobyte);
#endif
	write_start_bit();
	i2c_write_a_byte(i2caddr, WRITEBIT); //Address and write -bit
	i2c_write_a_byte(reg, PURE_BYTE);
	i2c_write_a_byte(hibyte, PURE_BYTE); //HI byte
	i2c_write_a_byte(lobyte, PURE_BYTE); //LO byte
	write_master_nack(); // No more bytes to be written
	write_stop_bit();
	return OK;
    }

/* --------------------------------------------------------------- */
u8 i2c_update_register(u8 i2caddr, u8 reg, u8 himask, u8 lomask, u8 hibyte, u8 lobyte)
{
	u8 hb, lb;

	// First read the current state of the register
	if (i2c_read_register(i2caddr, reg, &hb, &lb) == OK)
	{
#ifdef DEBUG
		printf("upd_reg/read:reg=0x%x,hi=0x%x,lo=0x%x\n", reg, hb, lb);
#endif
		// Apply data via masks
		hibyte |= hb & ~himask;
		lobyte |= lb & ~lomask;
		// Write updated register contents back to chip
#ifdef DEBUG
		printf("upd_reg/write:reg=0x%x,hi=0x%x,lo=0x%x\n", reg, hibyte, lobyte);
#endif
		return i2c_write_register(i2caddr, reg, hibyte, lobyte);
    }
	else return NOT_OK;
}
/* --------------------------------------------------------------- */
